namespace GossamerVR.Stripwise
{

    /*****
     * 
     * todo list:
     *      
     * add filter toggles in RestoreFromJSON for pending ignores, so that the disabled toggles are also preserved.
     * add a hierarchy listener to clothing containers to detect when clothing LIBRARY changed
     * add bounds box UI
     * enforce placing plugin on a person          
     *******/


    using System;
    using System.Collections;
    using System.Collections.Generic;
    using UnityEngine;
    using GPUTools.Cloth.Scripts;
    using System.Linq;
    using UnityEngine.UI;
    using SimpleJSON;

    public class StripwiseLite : MVRScript
    {
        public static class JSONKey
        {
            public const string Filter = "Filter";
        }
        public static class CutoffMode
        {
            public const string Off = "Off";
            public const string Normal = "Normal";
            public const string Inverted = "Inverted";            
        }
        private List<string> ModeChoices = new List<string>()
        {
            CutoffMode.Off, CutoffMode.Normal, CutoffMode.Inverted
        };

        /// <summary>
        /// how much to fade disabled clothing toggles by,
        /// set in a single place.
        /// </summary>
        private const float k_fade = 0.35f;

        private JSONStorableBool logJSON;

        // Main page elements
        private JSONStorableFloat
            thresholdJSON,
            multiplierJSON,
            falloffJSON,
            updateRateJSON;
        
        private JSONStorableFloat
            xMinJSON, xMaxJSON,
            yMinJSON, yMaxJSON,
            zMinJSON, zMaxJSON,
            hMinJSON, hMaxJSON,
            wMinJSON, wMaxJSON;


        private JSONStorableAction
            reloadActionJSON,
            loadIgnorePageActionJSON,
            loadMainPageActionJSON;

        private UIDynamicButton
            reloadClothingButton,
            loadIgnorePageButton,
            loadMainPageButton;

        private JSONStorableStringChooser 
            xModeJSON, yModeJSON, zModeJSON;

                
        private List<UIDynamic> mainPageElements = new List<UIDynamic>();
        private List<UIDynamic> filterPageElements = new List<UIDynamic>();

                

        private DAZCharacterSelector character;

        /// <summary>
        /// All known clothing bools stored in the character. populated on init and when resetting.
        /// We subscribe to these so we know if clothing has changed.
        /// Large libraries means a large list :(
        /// </summary>
        private List<JSONStorableBool> allClothingBools = new List<JSONStorableBool>();

        /// <summary>
        /// A list populated only by clothing that is or was active. These
        /// are the elements that get updated by the plugin.
        /// </summary>
        private List<ClothingEntry> trackedClothing = new List<ClothingEntry>();

        /// <summary>
        /// a 1 to 1 match of tracked clothing but only the uids.
        /// We do this instead of a dictionary because it's faster to iterate
        /// when we update frequently.
        /// </summary>
        private List<string> trackedUIDs = new List<string>();

        /// <summary>
        /// a list of dynamically created bools allowing the user to ignore specific clothing items.
        /// this is always in sync with the trackedUIDs list.
        /// </summary>
        private List<JSONStorableBool> filterJSONBools = new List<JSONStorableBool>();
        /// <summary>
        /// a list of dynamically created toggles 
        /// </summary>
        private List<UIDynamicToggle> filterToggles = new List<UIDynamicToggle>();

        /// <summary>
        /// we attend to clothing items that we're tracking
        /// that have been disabled. This way we can preserve ignore settings.
        /// </summary>
        private List<string> disabledUIDs = new List<string>();

        /// <summary>
        /// These are clothing items that have not yet found their settings content,
        /// presumably because they're still loading. They will be ignored until ready.
        /// </summary>
        private List<ClothingEntry> pendingClothing = new List<ClothingEntry>();

        /// <summary>
        /// These are UIDs of clothing we found in the scene JSON as being ignored.
        /// When a clothing item is detected, it is checked against this list
        /// when determining if it should be ignored or not, then removed from this list.
        /// </summary>
        private List<string> pendingIgnores = new List<string>();
        
        private Dictionary<ClothingEntry, float[]> rigidityCaches = new Dictionary<ClothingEntry, float[]>();
        private Dictionary<ClothingEntry, Vector3[]> positionCaches = new Dictionary<ClothingEntry, Vector3[]>();
        private bool initialized;
        private bool isDirty = true;
        private float commitSettingsTime = 0f;
        private float refreshPendingTime = 0f;

        private Dictionary<JSONStorableFloat, float> lastValueLookup = new Dictionary<JSONStorableFloat, float>();
        private Vector3 minVals, maxVals;

        private Bounds bounds = new Bounds();        
        private bool isQuitting = false;

        /// <summary>
        /// true if the filter page is shown. This ensures we 
        /// only activate a new toggle if we're on the right 'page'
        /// </summary>
        private bool showFilters = false;

        /// <summary>
        /// helper to iterate what side a filter toggle appears on.
        /// originally was a modulo of the count, but as the cap gets 
        /// reached the count doesn't change and they end up piling on one side.
        /// so instead we just switch this back and forth.
        /// </summary>
        private bool nextFilterToggleOnRight = true;

        public void OnEnable() { Log("OnEnable()"); }
        protected override void Awake()
        {
            base.Awake();
            Log("Awake()");
        }
        public override void Init()
        {
            base.Init();
            this.Log("Init()");

            /// MAIN PAGE ===========================================================================================================
            this.thresholdJSON = new JSONStorableFloat("Max Sim Value Affected", 1f, this.UpdateSettings, 0f, 1f);
            this.RegisterFloat(this.thresholdJSON);

            this.multiplierJSON = new JSONStorableFloat("Constant Multiplier", 1f, this.UpdateSettings, 0f, 3f, constrain:false);
            this.RegisterFloat(this.multiplierJSON);

            this.xMinJSON = new JSONStorableFloat("X Cutoff Min", -1f, this.UpdateSettings, -1f, 1f, constrain: false);
            this.RegisterFloat(this.xMinJSON);

            this.xMaxJSON = new JSONStorableFloat("X Cutoff Max", 1f, this.UpdateSettings, -1f, 1f, constrain: false);
            this.RegisterFloat(this.xMaxJSON);

            this.yMinJSON = new JSONStorableFloat("Y Cutoff Min", 0f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.yMinJSON);

            this.yMaxJSON = new JSONStorableFloat("Y Cutoff Max", 2.5f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.yMaxJSON);

            this.zMinJSON = new JSONStorableFloat("Z Cutoff Min", -1f, this.UpdateSettings, -1f, 1f, constrain: false);
            this.RegisterFloat(this.zMinJSON);

            this.zMaxJSON = new JSONStorableFloat("Z Cutoff Max", 1f, this.UpdateSettings, -1f, 1f, constrain: false);
            this.RegisterFloat(this.zMaxJSON);

            this.hMinJSON = new JSONStorableFloat("Bottom Effect Bounds", 0f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.hMinJSON);

            this.hMaxJSON = new JSONStorableFloat("Top Effect Bounds", 2.5f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.hMaxJSON);

            this.wMinJSON = new JSONStorableFloat("Left Effect Bounds", -1.25f, this.UpdateSettings, -1.25f, 1.25f, constrain: false);
            this.RegisterFloat(this.wMinJSON);

            this.wMaxJSON = new JSONStorableFloat("Right Effect Bounds", 1.25f, this.UpdateSettings, -1.25f, 1.25f, constrain: false);
            this.RegisterFloat(this.wMaxJSON);

            this.logJSON = new JSONStorableBool("Debug Log", false);
            this.RegisterBool(this.logJSON);

            this.falloffJSON = new JSONStorableFloat("Falloff", 0.5f, this.UpdateSettings, 0.01f, 4f, constrain: true);
            this.RegisterFloat(this.falloffJSON);

            this.updateRateJSON = new JSONStorableFloat("Update Rate", 20f, 0f, 90f, constrain: false);
            this.RegisterFloat(this.updateRateJSON);

            this.xModeJSON = new JSONStorableStringChooser("X Cutoff Mode", this.ModeChoices, CutoffMode.Normal, "X Cutoff Mode", this.OnChangeCutoffMode);            
            this.yModeJSON = new JSONStorableStringChooser("Y Cutoff Mode", this.ModeChoices, CutoffMode.Normal, "Y Cutoff Mode", this.OnChangeCutoffMode);            
            this.zModeJSON = new JSONStorableStringChooser("Z Cutoff Mode", this.ModeChoices, CutoffMode.Normal, "Z Cutoff Mode", this.OnChangeCutoffMode);
            this.RegisterStringChooser(xModeJSON);
            this.RegisterStringChooser(yModeJSON);
            this.RegisterStringChooser(zModeJSON);

            this.reloadActionJSON = new JSONStorableAction("Reload Clothing", this.ReloadClothing);
            this.RegisterAction(this.reloadActionJSON);

            this.loadIgnorePageActionJSON = new JSONStorableAction("Load Clothing Filters", this.LoadFilterPage);
            this.loadMainPageActionJSON = new JSONStorableAction("Load Main Page", this.LoadMainPage);


            /// IGNORE OPTIONS PAGE ===========================================================================================================
            
        }
        


        public override void InitUI()
        {
            base.InitUI();
            if (this.UITransform != null)
            {
                this.Log("InitUI");
                this.character = this.containingAtom.transform.Find(PersonUtil.Path.Geometry).GetComponent<DAZCharacterSelector>();

                this.InitMainUI();
                this.InitIgnoreUI();
                this.filterPageElements.ForEach(x => x.gameObject.SetActive(false));
                
                this.LoadAllClothing();
                this.initialized = true;
                
            }
            
        }

        private void InitMainUI()
        {
            // we'll need to access some of the UI elements briefly.
            // to reduce resource cost we'll use the same variable for all of them.
            UIDynamic tempElement_;
            UIDynamicButton tempButton_;
            /// LEFT SIDE =======================================
            bool isRight = false;

            this.mainPageElements.Add(this.CreateSlider(this.thresholdJSON, isRight));
            this.mainPageElements.Add(this.CreateSlider(this.multiplierJSON, isRight));
            this.mainPageElements.Add(this.CreateSlider(this.falloffJSON, isRight));

            tempElement_ = this.CreateSpacer(isRight);
            tempElement_.height = 10f;
            this.mainPageElements.Add(tempElement_);

            tempElement_ = this.CreateSlider(this.xMinJSON, isRight);
            this.SetUIColors(tempElement_, Color.red, 0.2f);
            this.mainPageElements.Add(tempElement_);

            tempElement_ = this.CreateSlider(this.yMinJSON, isRight);
            this.SetUIColors(tempElement_, Color.green, 0.2f);
            this.mainPageElements.Add(tempElement_);

            tempElement_ = this.CreateSlider(this.zMinJSON, isRight);
            this.SetUIColors(tempElement_, Color.blue, 0.2f);
            this.mainPageElements.Add(tempElement_);

            this.mainPageElements.Add(this.CreateSlider(this.hMinJSON, isRight));
            this.mainPageElements.Add(this.CreateSlider(this.wMinJSON, isRight));

            this.mainPageElements.Add(this.CreateSlider(this.updateRateJSON, isRight));



            /// RIGHT SIDE =======================================
            isRight = true;

            this.loadIgnorePageButton = this.CreateButton("Clothing Filters >>", isRight);
            this.loadIgnorePageActionJSON.RegisterButton(this.loadIgnorePageButton);
            this.mainPageElements.Add(this.loadIgnorePageButton);
            
            this.mainPageElements.Add(this.CreateScrollablePopup(this.xModeJSON, isRight));
            this.mainPageElements.Add(this.CreateScrollablePopup(this.yModeJSON, isRight));
            this.mainPageElements.Add(this.CreateScrollablePopup(this.zModeJSON, isRight));

            tempElement_ = this.CreateSpacer(isRight);
            tempElement_.height = 10f;
            this.mainPageElements.Add(tempElement_);

            tempElement_ = this.CreateSlider(this.xMaxJSON, isRight);
            this.SetUIColors(tempElement_, Color.red, 0.2f);
            this.mainPageElements.Add(tempElement_);

            tempElement_ = this.CreateSlider(this.yMaxJSON, isRight);
            this.SetUIColors(tempElement_, Color.green, 0.2f);
            this.mainPageElements.Add(tempElement_);

            tempElement_ = this.CreateSlider(this.zMaxJSON, isRight);
            this.SetUIColors(tempElement_, Color.blue, 0.2f);
            this.mainPageElements.Add(tempElement_);

            this.mainPageElements.Add(this.CreateSlider(this.hMaxJSON, isRight));
            this.mainPageElements.Add(this.CreateSlider(this.wMaxJSON, isRight));

            this.mainPageElements.Add(this.CreateToggle(this.logJSON, isRight));

            tempButton_ = this.CreateButton("Reload Clothing", isRight);
            this.reloadActionJSON.RegisterButton(tempButton_, isRight);
            this.mainPageElements.Add(tempButton_);
            this.SetUIColors(tempButton_, Color.yellow, 0.3f);
        }

        private void InitIgnoreUI()
        {
            UIDynamic tempElement_;
            UIDynamicButton tempButton_;
            /// LEFT SIDE =======================================
            bool isRight = false;

            this.loadMainPageButton = this.CreateButton("<< Main Page", isRight);
            this.loadMainPageActionJSON.RegisterButton(this.loadMainPageButton);
            this.filterPageElements.Add(this.loadMainPageButton);


            /// RIGHT SIDE =======================================
            isRight = true;
        }


        private void SetUIColors(Component target, Color tint, float amount)
        {
            if (target != null)
            {
                Image[] images = target.GetComponentsInChildren<Image>();
                foreach (var image in images)
                {
                    image.color = Color.Lerp(image.color, tint, amount);
                }
            }
        }


        private void LoadMainPage()
        {
            this.filterPageElements.ForEach(x=> x.gameObject.SetActive(false));
            this.mainPageElements.ForEach(x => x.gameObject.SetActive(true));
            this.showFilters = false;
        }
        private void LoadFilterPage() 
        {
            this.filterPageElements.ForEach(x => x.gameObject.SetActive(true));
            this.mainPageElements.ForEach(x => x.gameObject.SetActive(false));
            this.showFilters = true;
        }


        /// <summary>
        /// before we write to file, we convert our pose list to JSON and store it
        /// </summary>        
        public override JSONClass GetJSON(bool includeAppearance, bool includePhysical, bool forceStore)
        {
            this.Log("GetJSON()...");
            var json = base.GetJSON(includeAppearance, includePhysical, forceStore);
            var ignoreArray = new JSONArray();

            // we save anything we have ignored into the scene file.
            for (int i = 0; i < this.trackedUIDs.Count; i++)
            {
                if (!this.filterJSONBools[i].val)
                {
                    this.Log("... adding ignore: " + this.trackedUIDs[i]);
                    ignoreArray.Add(this.trackedUIDs[i]);
                }
            }

            json[JSONKey.Filter] = ignoreArray;

            return json;
        }

        public override void RestoreFromJSON(JSONClass json, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(json, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            this.Log("RestoreFromJSON()...");
            try
            {
                // when loading the scene, we populate a string list with 
                // the uids of the clothing we are ignoring.
                // once those items have loaded, we apply those filter settings.
                if (json.HasKey(JSONKey.Filter))
                {
                    JSONArray ignoreArray = json[JSONKey.Filter].AsArray;
                    int index;
                    foreach (var obj in ignoreArray)
                    {
                        string uid = ((JSONNode)obj).Value;

                        
                        index = this.trackedUIDs.IndexOf(uid);
                        if (index >= 0) 
                        {
                            // if we're already tracking the clothes, ignore them immediately.
                            this.filterToggles[index].toggle.isOn = false;
                        }
                        else
                        {
                            // otherwise cache the uid to ignore when it loads.
                            this.pendingIgnores.Add(uid);
                        }
                        
                    }
                }
                else
                {
                    this.Log("... no filter.");
                }
            }
            catch (Exception e)
            {
                this.Log("StripwiseLite failed to load filtered uids from JSON.", isError: true);
            }
            this.isDirty = true;
        }



        /// <summary>
        /// subscirbes to all clothing JSON bools and tracks any new active clothes.
        /// </summary>
        private void LoadAllClothing()
        {
            this.Log("LoadAllClothing()");
            this.allClothingBools.Clear();
            IEnumerable<string> clothingBoolNames = this.character.GetBoolParamNames().Where(x=> x.StartsWith("clothing:"));
            JSONStorableBool jsonBool;
            string clothingUID;

            foreach(string boolName in clothingBoolNames)
            {
                jsonBool = this.character.GetBoolJSONParam(boolName);                
                jsonBool.setJSONCallbackFunction -= this.OnToggleClothingActive;
                jsonBool.setJSONCallbackFunction += this.OnToggleClothingActive;
                this.allClothingBools.Add(jsonBool);                
                if (jsonBool.val)
                {                    
                    // if the clothing is active, we track it so we'll update it.
                    clothingUID = boolName.Substring(9);                    
                    this.AddClothingEntry(clothingUID);
                }
            }
            this.isDirty = true;
        }

        /// <summary>
        /// Subscribed in LoadAllClothing to every clothing's associated state bool.
        /// Any time a clothing item is toggled on or off, this gets called
        /// </summary>        
        private void OnToggleClothingActive(JSONStorableBool jsonBool)
        {
            this.Log("Clothing toggled: " + jsonBool.name);
            string uid = jsonBool.name.Substring(9);
            if (jsonBool.val)
            {
                // if we're enabling clothing, we ensure it's tracked
                this.Log("...activating");
                this.disabledUIDs.Remove(uid);
                this.AddClothingEntry(uid);
            }
            else 
            {
                // we cache a limited number of disabled items 
                // that were previously worn. This way we preserve
                // any ignore settings the user applied to them.
                this.Log("...deactivating");
                this.disabledUIDs.Add(uid);

                // if we have a toggle for the clothing, gray it out when it's disabled.
                UIDynamicToggle filterUI = this.GetTrackedFeature(this.filterToggles, uid);
                if (filterUI != null) this.SetUIColors(filterUI.toggle, Color.clear, k_fade);

                if (this.disabledUIDs.Count > 20)
                {
                    // we don't track more than 20 disabled items at a time.
                    // new ones will replace the oldest.
                    this.RemoveClothingEntry(this.disabledUIDs[0]);
                }
            }
            this.isDirty = true;
        }

        private void OnChangeCutoffMode(JSONStorableStringChooser jsonChooser)
        {
            this.isDirty = true;
        }

        /// <summary>
        /// Adds and tracks a new clothing entry if the uid is not already tracked.
        /// </summary>        
        private void AddClothingEntry(string uid)
        {
            if (!this.trackedUIDs.Contains(uid))
            {
                DAZClothingItem item = this.character.GetClothingItem(uid);
                var entry = new ClothingEntry(item);
                if (entry.IsSimulated)
                {
                    this.Log("Tracking clothing: " + uid);
                    this.trackedUIDs.Add(uid);
                    this.trackedClothing.Add(entry);
                    entry.Ignore = this.pendingIgnores.Contains(uid);
                    this.AddClothingFilterToggle(uid, item.displayName, !entry.Ignore);
                    foreach(var ignored in this.pendingIgnores)
                    {
                        this.Log("ignored includes: " + ignored);
                    }
                    this.pendingIgnores.Remove(uid);

                    if (!entry.IsValid)
                    {
                        this.Log("...clothing has not yet been assigned simulation settings, adding to pending list.");
                        this.pendingClothing.Add(entry);
                    }
                }
                else
                {
                    this.Log("Ignoring non-simulated clothing: " + uid);
                }
            }
            else
            {
                this.Log("Clothing already tracked: " + uid);
                this.SetUIColors(this.GetTrackedFeature(this.filterToggles, uid)?.toggle, Color.white, 1f);
            }            
        }

        /// <summary>
        /// Removes all tracking for the uid and its associated clothing entry.
        /// </summary>
        /// <param name="uid"></param>
        private void RemoveClothingEntry(string uid)
        {
            int index = this.trackedUIDs.IndexOf(uid);
            if (index >= 0)
            {
                var entry = this.trackedClothing[index];
                this.RemoveClothingFilterToggle(uid);
                this.pendingClothing.Remove(entry);
                this.trackedClothing.RemoveAt(index);
                this.trackedUIDs.RemoveAt(index);
                this.RemoveClothingFilterToggle(uid);
                this.disabledUIDs.Remove(uid);
            }
            this.disabledUIDs.Remove(uid);
            this.isDirty = true;
            
        }

        private ClothingEntry GetClothingEntry(string uid)
        {
            int index = this.trackedUIDs.IndexOf(uid);
            if (index >= 0) return this.trackedClothing[index];
            return null;
        }
        
        /// <summary>
        /// Provides a generic way of using a uid as a lookup 
        /// for a given list. 
        /// </summary>        
        private T GetTrackedFeature<T>(List<T> list, string uid)
        {
            int index = this.trackedUIDs.IndexOf(uid);
            if (index >= 0 && index < list.Count) return list[index];
            else return default(T);            
        }

        private void AddClothingFilterToggle(string uid, string name, bool initialValue)
        {
            JSONStorableBool json = new JSONStorableBool(name, initialValue, this.OnToggleClothingFilter);
            
            // we don't want to save these because they're generated dynamically.
            json.isStorable = false;
            json.isRestorable = false;

            this.filterJSONBools.Add(json);
            this.RegisterBool(json);
            bool isOnRight = this.filterToggles.Count < 20 ? this.filterToggles.Count % 2 != 0 : this.nextFilterToggleOnRight;
            var filterUI = this.CreateToggle(json, isOnRight);
            this.nextFilterToggleOnRight = !this.nextFilterToggleOnRight;
            this.filterPageElements.Add(filterUI);
            this.filterToggles.Add(filterUI);

            // if the clothing item itself is inactive, gray out the toggle, but still allow its use.
            if (!this.GetClothingEntry(uid)?.IsActive ?? true) 
            {
                this.SetUIColors(filterUI.toggle, Color.clear, k_fade);
            }
            else
            {
                this.SetUIColors(filterUI.toggle, Color.white, 1f);
            }

            if (!this.showFilters) filterUI.gameObject.SetActive(false);
        }

        private void RemoveClothingFilterToggle(string uid)
        {
            int index = this.trackedUIDs.IndexOf(uid);
            if (index >= 0)
            {                
                JSONStorableBool json = this.filterJSONBools[index];
                this.DeregisterBool(json);
                UIDynamicToggle toggle = this.filterToggles[index];
                this.filterPageElements.Remove(toggle);
                this.filterJSONBools.Remove(json);
                this.filterToggles.Remove(toggle);
                json.setJSONCallbackFunction = null;
                json.toggle = null;
                Destroy(toggle.gameObject);
            }
        }

        private void OnToggleClothingFilter(JSONStorableBool jsonBool)
        {
            int index = this.filterJSONBools.IndexOf(jsonBool);
            if (index >= 0)
            {
                var entry = this.trackedClothing[index];
                entry.Ignore = !jsonBool.val;
                
                if (entry.IsReady && entry.Ignore)
                {
                    entry.Settings.Reset();
                }
            }
            this.isDirty = true;
        }

        /// <summary>
        /// 
        /// </summary>
        private void ClearAll(bool clearIgnores = true)
        {
            if (!this.isQuitting)
            {
                this.pendingIgnores.Clear();
                foreach (ClothingEntry entry in this.trackedClothing)
                {
                    if (entry.Settings != null) entry.Settings.Reset();

                }

                for (int i =0; i < this.trackedUIDs.Count; i++)
                {
                    if (!clearIgnores && this.trackedClothing[i].Ignore)
                    {
                        // if we want to preserve active ignores, such as when clicking ReloadClothing,
                        // then we add currently ignored items to 'pending' so they come back ignored.
                        this.pendingIgnores.Add(this.trackedUIDs[i]);
                    }

                    this.RemoveClothingEntry(this.trackedUIDs[i]);
                    i--;
                }
                
                this.rigidityCaches.Clear();
                this.positionCaches.Clear();

                if (this.logJSON.val)
                {
                    Log($"ClearAll() called, these should all be ZERO: " +
                        $"disabled: {disabledUIDs.Count}, entries: {trackedClothing.Count}, " +
                        $"uids: {trackedUIDs.Count}, pending: {pendingClothing.Count}, " +
                        $"toggles: {filterToggles.Count}, bools: {filterJSONBools.Count}");                    
                }
                this.isDirty = true;
            }
        }

       
        /// <summary>
        /// Clears all tracked values and reloads all clothing.
        /// Similar to reloading the plugin but you keep all your settings.
        /// This is necessary if the library changes, so we can resubscribe
        /// </summary>
        private void ReloadClothing()
        {
            this.ClearAll(clearIgnores:false);
            this.LoadAllClothing();
        }



        /// <summary>
        /// Callback for when a slider or UI element changes.
        /// This will set the dirty flag if the change was significant enough to warrant it.
        /// this significance threshold helps allow for timeline to animate sliders
        /// without triggering a refresh if they don't change.
        /// </summary>
        private void UpdateSettings(JSONStorableFloat changedFloat)
        {            
            if (!this.lastValueLookup.ContainsKey(changedFloat))
            {
                this.lastValueLookup[changedFloat] = changedFloat.val;
                this.isDirty = true;
            }
            else if (Mathf.Abs(this.lastValueLookup[changedFloat] - changedFloat.val) > 0.01f)
            {
                this.isDirty = true;
            }

            if (this.isDirty)
            {
                this.bounds.min = new Vector3(this.wMinJSON.val, this.hMinJSON.val, -2f);
                this.bounds.max = new Vector3(this.wMaxJSON.val, this.hMaxJSON.val, 2f);
            }

        }


        private void Update()
        {
            if (this.initialized && this.trackedClothing.Count > 0)
            {
                this.commitSettingsTime += Time.deltaTime;
                this.refreshPendingTime += Time.deltaTime;
                if (this.pendingClothing.Count > 0 && this.refreshPendingTime >= 1f)
                {
                    this.RefreshPendingClothing();
                    this.refreshPendingTime = 0f;
                }

                if (this.isDirty && (this.updateRateJSON.val < 1f || this.commitSettingsTime * this.updateRateJSON.val > 1f))
                {
                    this.isDirty = false;
                    this.CalculateAndCommitProperties();                    
                    this.commitSettingsTime = 0f;
                }
            }
        }

        /// <summary>
        /// Tries to load any missing settings on clothing entries.
        /// removes them from the pending list if successful.
        /// </summary>
        private void RefreshPendingClothing()
        {
            for (int i = 0; i < this.pendingClothing.Count; i++)
            {
                ClothingEntry entry = this.pendingClothing[i];
                string uid = entry.Clothing.uid;
                if (entry.Clothing.active)
                {
                    if (!entry.LoadSettings())
                    {
                        // if our load attempt failed then we have timed out, and we abandon the item.
                        // this shouldn't happen but may under the following conditions: (note to self:  account for them if it does happen)
                        //      - the clothing item is destroyed but we are still tracking (including in pending) the entry for it.
                        //      - the clothing item is labeled as sim / custom but is not actually simulated and has no settings.
                        //      - the clothing type (sim/wrap) has changed resulting in a removal of the settings
                        //      - the 'custom' type does not always have settings.
                        this.RemoveClothingEntry(uid);
                        i--;
                        this.isDirty = true;
                        this.Log(uid + " has been pending simulation settings for too long and will be ignored." +
                                    "Please let GossamerVR know, as any non-sim clothes shouldn't be pending.", isError:true);

                    }
                    else if (entry.IsReady)
                    {                        
                        this.Log(uid + " Sim settings have now been applied. Removing from pending list.");
                        this.pendingClothing.RemoveAt(i);
                        i--;
                        this.isDirty = true;
                    }
                }
            }
        }

        /// <summary>
        /// Does the work of iterating the clothing and their vertices,
        /// applying the changes to the clothing simulation.
        /// </summary>
        private void CalculateAndCommitProperties()
        {
            
            /// since we're iterating tens (sometimes hundreds) of thousands of vertices,
            /// we cache a lot of JSON properties to scrape just a tiny bit more performance,
            /// as well as cache the results from some string comparisons
            this.minVals.x = this.xMinJSON.val;
            this.minVals.y = this.yMinJSON.val;
            this.minVals.z = this.zMinJSON.val;

            this.maxVals.x = this.xMaxJSON.val;
            this.maxVals.y = this.yMaxJSON.val;
            this.maxVals.z = this.zMaxJSON.val;

            float cutoffPower = this.falloffJSON.val;
            float multiplier = this.multiplierJSON.val;
            
            bool ignoreX = this.xModeJSON.val == CutoffMode.Off;
            bool ignoreY = this.yModeJSON.val == CutoffMode.Off;
            bool ignoreZ = this.zModeJSON.val == CutoffMode.Off;

            float dirX = this.xModeJSON.val == CutoffMode.Inverted ? -1 : 1;
            float dirY = this.yModeJSON.val == CutoffMode.Inverted ? -1 : 1;
            float dirZ = this.zModeJSON.val == CutoffMode.Inverted ? -1 : 1;

            
            float[] rigidityCache = null;
            Vector3[] positionCache;
            ClothingEntry entry;
            ClothSettings sim;
            for (int i = 0; i < this.trackedClothing.Count; i++)
            {
                entry = this.trackedClothing[i];
                if (entry.Ignore || !entry.IsValid || !entry.IsSimActive)
                {
                    // skip tracked clothing toggled off in the filter page, as well as any that are disabled or not simulated.
                    // we don't strip these from the tracked clothing because some of them are still loading their sim settings,
                    // and others simply have their sim unchecked in physics settings, and we're waiting on the user to enable it.
                    continue;
                }

                
                sim = entry.Settings;

                if (!this.rigidityCaches.TryGetValue(entry, out rigidityCache))
                {
                    sim.Reset();
                    sim.BreakEnabled = false;
                }
                
                /// pull the simulation points so the GPU marshals the data for us to use.
                /// for safety, we do this after potentially calling Reset() but before iterating/caching.
                sim.Runtime.Particles.PullData();
                sim.Runtime.PointJoints.PullData();

                if (rigidityCache == null)
                {
                    // this will be null if we failed to fetch it a few lines before the data pull.
                    // in such a case we build a new cache.
                    this.Log("Stripwise caching new sim clothing: " + entry.Clothing.displayName);
                    this.CacheValues(entry, out rigidityCache, out positionCache);
                }
                else
                {
                    positionCache = this.positionCaches[entry];
                }


                // ========== Begin bulk vertex work ===============/
                int vertIndex = 0;
                float xRigid = 1f;
                float yRigid = 1f;
                float zRigid = 1f;
                float rigidity = 1f;
                Vector3 pos;                
                for (vertIndex = 0; vertIndex < sim.Runtime.PointJoints.Count; vertIndex++)
                {
                    pos = positionCache[vertIndex];
                    rigidity = rigidityCache[vertIndex];
                    if (rigidity <= this.thresholdJSON.val && this.bounds.Contains(pos))
                    {
                        xRigid = ignoreX ? 1f : Mathf.Pow(this.GetCutoffValue(pos.x, this.minVals.x, this.maxVals.x, dirX), cutoffPower);
                        yRigid = ignoreY ? 1f : Mathf.Pow(this.GetCutoffValue(pos.y, this.minVals.y, this.maxVals.y, dirY), cutoffPower);
                        zRigid = ignoreZ ? 1f : Mathf.Pow(this.GetCutoffValue(pos.z, this.minVals.z, this.maxVals.z, dirZ), cutoffPower);
                        rigidity *= multiplier * xRigid * yRigid * zRigid;
                    }
                    sim.Runtime.PointJoints.Data[vertIndex].Rigidity = rigidity;
                }

                // =============End bulk vertex work ================//


                /// push the updated simulation points back so the data is available for the GPU
                sim.Runtime.PointJoints.PushData();
                sim.Runtime.Particles.PushData();

            }
        }

        /// <summary>
        /// Creates a new rigidity and position cache for the clothing,
        /// adding it to the lookup and providing the results.
        /// </summary>        
        private void CacheValues(ClothingEntry entry, out float[] rigidityCache, out Vector3[] positionCache)
        {
            rigidityCache = new float[entry.Settings.Runtime.PointJoints.Count];
            positionCache = new Vector3[entry.Settings.Runtime.Particles.Count];
            this.rigidityCaches[entry] = rigidityCache;
            this.positionCaches[entry] = positionCache;
            for (int i = 0; i < entry.Settings.Runtime.PointJoints.Count; i++)
            {
                rigidityCache[i] = entry.Settings.Runtime.PointJoints.Data[i].Rigidity;
                positionCache[i] = entry.Settings.Runtime.Particles.Data[i].Position;
            }
        }

        /// <summary>
        /// Returns a value between 0 and 1 based on whether or not the input is between min and max.
        /// allows min and max to be reversed.
        /// </summary>                
        private float GetCutoffValue(float input, float min, float max, float direction)
        {            
            return Mathf.Clamp01( direction * ((input - min) * (max - input)));
        }


        private void OnDestroy()
        {
            foreach(var boolJSON in this.allClothingBools)
            {
                if (boolJSON != null) boolJSON.setJSONCallbackFunction -= this.OnToggleClothingActive;
            }
            this.ClearAll();
        }


        private void Log(string msg, bool isError = false)
        {
            if (((this.logJSON != null && this.logJSON.val) || isError) && SuperController.singleton != null)
            {
                if (isError)
                {
                    SuperController.LogError(msg);
                }
                else SuperController.LogMessage(msg);
            }

        }

        private void OnApplicationQuit()
        {
            this.isQuitting = true;
        }

    }
}